"""Causal Inference Toolbox

Sub-package dedicated to functionalities related to the field of causal inference. 
For a introduction on the subject see: https://mixtape.scunning.com/introduction.html#what-is-causal-inference
"""

from pyspark_ds_toolbox.causal_inference import ps_matching, diff_in_diff